#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>

#include "defs.h"
#include "math.h"
#include "templates.cpp"
#include "string.h"
#include "faces.h"

int uscent[4][4]={{-1,1,0,5},{1,-1,3,2},{0,3,-1,4},{5,2,4,-1}};

//Facce 'ff' un vertice 'pt' e una direzione 'dir' individuano uno spigolo.
//restituisce la terna data dai colori delle 3 facce incidenti
teta::teta(const std::vector<faccia>& ff, int pt, int dir) {
  elem[0]=0; elem[1]=0; elem[2]=0;
  int sz = ff.size();
  for (int j=0; j<sz; j++)
    for (int i=0; i<ff[j].lati(); i++)
      if (ff[j].get_passaggio(i) == pt) {
        int d = mod(ff[j].get_uscita(i) - ff[j].get_entrata(i),4);
        if (ff[j].get_uscita(i)==dir) elem[d-1]=ff[j].get_colore();
        if (ff[j].get_entrata(i)==dir) elem[4-d-1]=-ff[j].get_colore();
      }
}


//restituisce il tetraedro dato dalle facce 'ff' (e dai loro colori) 
//associato al vertice numero 'pt'
//se 'val_assoluto' == TRUE allora (alla fine) sostituisce elem[j] con abs(elem[j])
tetraedro::tetraedro(const std::vector<faccia>& ff, int pt) {
  for (int i=0; i<6; i++) elem[i]=0;
  int a[6];
  for (int j=0; j<ff.size(); j++)
    for (int i=0; i<ff[j].lati(); i++)
      if (ff[j].get_passaggio(i) == pt) 
				a[uscent[ff[j].get_uscita(i)][ff[j].get_entrata(i)]] = ff[j].get_colore();
  *this = tetraedro(a[5], a[1], a[3], a[4], a[2], a[0]);
}

//dice se c'e' un counterpass (cioe' teta == (a, -a, ?))
bool teta::counterpass() const {
  return (elem[0] == -elem[1] || elem[1] == -elem[2] || elem[2] == -elem[0]);
}

//dice se c'e' una faccia che passa due volte nello stesso senso,
//cioe' teta == (a, a, ?)
bool teta::pass_twice() const {
  return (elem[0] == elem[1] || elem[1] == elem[2] || elem[2] == elem[0]);
}

//se teta == (a, a, x) oppure teta == (a, -a, x) con x diverso da a e -a,
//restituisce x.
//Altrimenti restituisce 0.
int teta::faccia_sola() const {
  teta tt = abs(*this);
  for (int t=0; t<3; t++)
    if (tt.elem[t] == tt.elem[(t+1)%3] && tt.elem[t]!=tt.elem[(t+2)%3])
      return elem[(t+2)%3];
  return 0;
}

//dice se abs(elem[t]) e' lo stesso per t=0,1,2
bool teta::triple_counterpass() const {
  return (facce().size()==1);
}

//dice se elem[t]==0 per tutti i t, modulo 'modulo' se modulo!=0.
bool teta::nullo() const {
	return (prodotto()==0);
}

//dice se e' ammissibile (disuguaglianza triangolare + numero pari di elementi dispari)
bool teta::ammissibile() const {
  for (int i=0; i<3; i++) {
    int n=elem[i]+elem[(int) mod(i+1,3)]-elem[(int) mod(i+2,3)];
    if (n<0 || mod(n,2)) return 0;
  }
  return 1;
}

//dice se e' r-ammissibile (ammissibile + somma < 2r-4)
bool teta::ammissibile(int r) const {
  if (!ammissibile() || (elem[0]+elem[1]+elem[2]>2*r-4)) return 0;
  return 1;
}

//restituisce i valori assoluti degli elem[], ordinati e senza ripetizioni
//(quindi le facce adiacenti)
std::vector<int> teta::facce() const{
  std::vector<int> ls(0);
  for (int t=0; t<3; t++)
    ls.push_back(abs(elem[t]));
	elimina_duplicati(ls);
  return ls;
}

//restituisce la somma degli elem[t]
int tetraedro::somma() const {
	int s=0; 
	for(int i=0; i<6; i++) s+=elem[i]; 
	return s;
}

//restituisce il valore massimo di elem[j]
int tetraedro::max_value () const {
  int m=0;
  for (int i=0; i<6; i++) if (m<elem[i]) m=elem[i];
  return m;
}

//colora le facce secondo il vettore z
void colora(const std::vector<faccia>& ff, const std::vector<int> z) {
  for (int i=0; i<ff.size(); i++) ff[i].set_colore(z[i]);
}

//colora le facce secondo {1,2,3, ...}
void colora_progressivamente(const std::vector<faccia>& ff) {
  std::vector<int> zz(0);
  for (int i=0; i<ff.size(); i++)
    zz.push_back(i+1);
  colora(ff,zz);
}

//serve per trovare le superfici contenute. Ordina ogni faccia che passa dallo spigolo (vert,col) in
//modo che passaggio[0]==vert e uscita[0]==col
void ordina_da_spigolo(std::vector<faccia> ff, int vert, int col) {
  for (int j=0; j<ff.size(); j++) {
    faccia f = ff[j];
		bool finito = FALSE;
    const int lt = f.lati();
    for (int i=0; i<f.lati() && !finito; i++)
      if (f.get_passaggio(i) == vert) {       //ruota
        if (f.get_entrata(i) == col) {
          f.inverti(); i=lt-i-1;
        }
        if (f.get_uscita(i) == col) {
          f.ruota(i);
          finito = TRUE;
        }
      }
  }
}

//dice se la faccia e' embedded (funziona anche se non e' chiusa!)
bool faccia::embedded() const {
  return tutti_diversi(passaggio);
}

//restituisce il numero di lati della faccia
//(se non e' chiusa, in verita' e' il numero di vertici incontrati)
int faccia::lati() const {
  return passaggio.size();
}

//inverte l'ordine dei vertici (trasformando quindi entrate in uscite e viceversa)
void faccia::inverti() {
	std::swap_ranges(uscita.begin(), uscita.end(), entrata.begin());
	std::reverse(entrata.begin(), entrata.end());
	std::reverse(uscita.begin(), uscita.end());
	std::reverse(passaggio.begin(), passaggio.end());
}

//ruota una faccia di n
void faccia::ruota(int n) {
  std::rotate(passaggio.begin(), passaggio.begin() + n, passaggio.end());
  std::rotate(entrata.begin(), entrata.begin() + n, entrata.end());
  std::rotate(uscita.begin(), uscita.begin() + n, uscita.end());
}

//aggiunge il passaggio p, con uscita u ed entrata e
void faccia::add_passaggio(int p, int u, int e) {
  passaggio.push_back(p); uscita.push_back(u); entrata.push_back(e);
}

//-teta e' un segno meno membro a membro
teta operator-(teta th) {
  return teta(-th.elem[0], -th.elem[1], -th.elem[2]);
}

//ruota di n
void teta::ruota(int n) {
  n = n%3;
  std::rotate(&elem[0],&elem[n],&elem[3]);
}

//dice se i due lati determinanti t1 e t2 contengono un loop di lunghezza 2
//bordante un disco
//ovvero, dice se (a meno di rotazione o di moltiplicazione per -1 di t1 o t2) 
//abbiamo t1 = [a,b,c], t2 = [a,b,d] con a!=b
bool short_loop(teta t1, teta t2) {
  for (int caso=0; caso<=1; caso++) {
    if (caso) {
			t2 = -t2;
			std::swap (t2.elem[0], t2.elem[1]);
		}
    for (int rotaz=0; rotaz<3; rotaz++) {
      for (int rotaz2=0; rotaz2<3; rotaz2++) {
        if (t1.elem[0] != t1.elem[1] && t1.elem[0] == t2.elem[0] && t1.elem[1] == t2.elem[1])
          return TRUE;
        t2.ruota(1);
      }
      t1.ruota(1);
    }
  }
  return FALSE;
}

//due teta sono uguali se lo sono come insiemi (non ordinati)
bool operator==(teta t1, teta t2) {
  t1.ordina(); t2.ordina();
  for (int t=0; t<3; t++)
    if (t1.elem[t] != t2.elem[t])
      return FALSE;
  return TRUE;
}

//restituisce un teta con i valori assoluti degli elem[]
teta abs(teta t) {
  return teta(abs(t.elem[0]), abs(t.elem[1]), abs(t.elem[2]));
}

//scrive su f
void teta::scrivi(std::ostream& f) const {
  f << "{";
  for (int t=0; t<3; t++) {
    f << elem[t];
    if (t<2) f << ",";
  }
  f << "}";
}

//restituisce un tetraedro con gli elem[] calcolati modulo 'modulo'
tetraedro operator% (const tetraedro& t, int modulo) {
  tetraedro ris = tetraedro();
  for (int i=0; i<6; i++)
    ris.elem[i] = t.elem[i] % modulo;
  return ris;
}

//restituisce un teta con gli elem[] calcolati modulo 'modulo'
teta operator% (const teta& t, int modulo) {
  teta ris = teta();
  for (int i=0; i<3; i++)
    ris.elem[i] = t.elem[i] % modulo;
  return ris;
}
