#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <list>
#include <vector>
#include <algorithm>

#include "defs.h"
#include "math.h"
#include "string.h"
#include "templates.cpp"
#include "ograph.h"
#include "faces.h"
#include "filters.h"
#include "io.h"

typedef std::list<spigolo_e_tor>::iterator st_iterator;
typedef std::list<spigolo_e_tor>::const_iterator st_c_iterator;
typedef std::list<std::list<spigolo_e_tor> >::iterator f_iterator;
typedef std::list<std::list<spigolo_e_tor> >::const_iterator f_c_iterator;
typedef std::list<spigolo_e_tor> filtro;

extern int Aquattro[12][4];
extern int RR[4];

int max_cicli = 5;
int max_paths = 4;

bool usa_orbite = 1;
bool usa_occhiali = 1;
bool usa_scorpione = 1;
bool usa_cappello = 1;
bool usa_maschera = 1;
bool usa_cicli = 1;
bool usa_path = 1;
bool usa_quadrato = 1;
bool usa_casa = 1;
int scrivi_filtri = 0;    //se 0 non scrive, se >0 scrive filtri di lunghezza <= scrivi_filtri

//ordine tra spigolo_e_tor
bool operator<(const spigolo_e_tor& st1, const spigolo_e_tor& st2) {
	return st1.spigolo < st2.spigolo;
}

//aggiunge un filtro piazzandolo al posto giusto
void aggiungi(std::vector<std::vector<std::list<filtro> > >& ris, filtro nuovo) {
	if (scrivi_filtri && nuovo.size()<=scrivi_filtri) scrivi(nuovo);
	if (nuovo.empty()) return;
	st_iterator massimo = max_element(nuovo.begin(), nuovo.end());
	int spig = massimo->spigolo;
	int tors = massimo->tor.get_t();
	nuovo.erase(massimo);
	ris[spig][tors].push_back(nuovo);
}

std::vector<std::vector<std::list<filtro> > > filtri_brick (const igrafo& ig) {
	int l = ig.lung();
	std::list<filtro> vuota;
	std::vector<std::vector<std::list<filtro> > > ris(2*l, std::vector<std::list<filtro> >(4, vuota));
	tensor2 n_spigolo = tensor2(l, std::vector<int>(4,0));
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      n_spigolo[i][j] = ig.v2spigolo(i,j);
	
	//cappello: chiediamo l>=6, cosi' C_{2,l-2} non e' un mattone. I C_{2,l-2} per l<6 sono infatti
	//gli unici mattoni con un 2-ponte che collega un cappello ed una catena, ed il teorema si applica
	//alla catena ma non al cappello. In tutti gli altri casi il teorema si applica ad ogni cappello.
  if (usa_cappello && l>=6) for (int i=0; i<l; i++) if (ig.tipo(i) == 0) 
    for (int j1=0; j1<3; j1++)
      for (int j2=j1+1; j2<4; j2++)
				if (ig.get_v(i,j1)>i && ig.get_v(i,j1)==ig.get_v(i,j2)) {
					int i2 = ig.get_v(i,j1);
					if (ig.tipo(i2) == 0) for (int k=0; k<4; k++)
						if (k!=j1 && k!=j2) for (int h=0; h<4; h++)
							if (ig.get_v(i,k) == ig.get_v(i2,h)) {
								std::vector<torsione> tor(4,0); int spigolo[4]; std::vector<int> c(2,0);
								::cappello(j1,ig.get_col(i,j1),j2,ig.get_col(i,j2),k,ig.get_col(i,k),h,
													 ig.get_col(i2,h),tor,c);

								//ci sono due tipi di cappello, nel primo caso viene filtrato tutto:
								if (mod(c[0],2) == mod(c[1],2)) {
									for (int t=0; t<4; t++) {
										filtro nuovo(1,spigolo_e_tor(0,t));
										aggiungi(ris, nuovo);
									}
									return ris;
								}

								//nell'altro caso, gli unici valori ammessi sono 3222 e 2311. Ovvero:
								//torsioni singole escluse:
								spigolo_e_tor esclusi[6] = {(0,0),(0,1),(1,0),(1,1),(2,0),(3,0)};
								spigolo[0] = n_spigolo[i][j1];
								spigolo[1] = n_spigolo[i][j2];
								spigolo[2] = n_spigolo[i][k];
								spigolo[3] = n_spigolo[i2][h];
								if (c[0]==3) std::swap(spigolo[0], spigolo[1]);
								for (int t=0; t<6; t++) {
									int spig = spigolo[esclusi[t].spigolo];
									torsione tors = esclusi[t].tor + tor[esclusi[t].spigolo];
									filtro nuovo(1,spigolo_e_tor(spig,tors));
									aggiungi(ris, nuovo);
								}
								//coppie di torsioni escluse:
								spigolo_e_tor esclusi2[4][2] = {{(0,3),(2,1)},{(0,3),(3,1)},{(1,3),(2,2)},{(1,3),(3,2)}};
								for (int t=0; t<4; t++) {
									int spig[2]; torsione tors[2];
									for (int t2=0; t2<2; t2++) {
										spig[t2] = spigolo[esclusi2[t][t2].spigolo];
										tors[t2] = esclusi2[t][t2].tor + tor[esclusi2[t][t2].spigolo];
									}
									filtro nuovo;
									for (int t=0; t<2; t++) nuovo.push_back(spigolo_e_tor(spig[t], tors[t]));
									aggiungi(ris, nuovo);
								}
								
                //inoltre, gli spigoli sul 2-ponte non possono avere valore 3:
								int i3=ig.get_v(i,k);
								for (int t=0; t<4; t++) {
									int spig = n_spigolo[i3][t];
									filtro nuovo(1,spigolo_e_tor(spig,3));
									aggiungi(ris, nuovo);
								}
							}
				}

  //cicli
	if (scrivi_filtri) std::cout << "Cicli: " << std::endl;
	tensor3 cic = cicli(ig, max_cicli);
  if (usa_cicli) for (int lati = 2; lati<=max_cicli && lati<=l; lati++) {  	
		int max = cic[lati].size();
    for (int c=0; c<max; c++) {
			std::vector<int> vert = cic[lati][c];
      std::vector<int> ent(lati,0); std::vector<int> usc(lati,0);
      set_ciclo(ent, usc, vert, 0, lati, ig, ris, n_spigolo);
			//se ci sono due lati, gli spigoli sul 2-ponte non possono avere valore 3:
      if (l>3 && lati == 2) for (int t=0; t<2; t++) for (int t2=0; t2<4; t2++)
				if (ig.get_v(vert[t], t2)!=vert[1-t]) {
					int spig = n_spigolo[vert[t]][t2];
					filtro nuovo(1, spigolo_e_tor(spig, 3));
					aggiungi(ris, nuovo);
				}
    }
  }

  //maschere (Filtro 5)
	if (scrivi_filtri) std::cout << "Maschere: " << std::endl;
  if (usa_maschera) for (int i=0; i<l; i++) if (ig.tipo(i)==1) {
		std::vector<int> v(4,0);
		for (int j=0; j<4; j++) v[j] = ig.get_v(i,j);
		int spigolo[4]; torsione tor[4];
		for (int t=0; t<4; t++) {
			int t2=0;
			for (; t==t2 || v[t]!=v[t2]; t2++) {}
			tor[t]=::maschera(t, ig.get_col(i,t), t2, ig.get_col(i,t2));
			filtro nuovo(1, spigolo_e_tor(n_spigolo[i][t], tor[t]));
			aggiungi(ris, nuovo);
		}
		for (int t=1; t<4; t++) {
			if (v[t]!=v[0]) {
				int spig[2]; spig[0]=n_spigolo[i][0]; spig[1]=n_spigolo[i][t];
				for (int t2=0; t2<2; t2++) {
					torsione tors[2]; 
					tors[0] = tor[0] + 1+t2; 
					tors[1] = tor[t] + (2-t2);
					filtro nuovo;
					for (int k=0; k<2; k++)
						nuovo.push_back(spigolo_e_tor(spig[k], tors[k]));
					aggiungi(ris, nuovo);
				}
			}
		}
	}

	//paths
	if (scrivi_filtri) std::cout << "Paths: " << std::endl;
	std::vector<int> vert(l+1,0);
  std::vector<int> ent(l+1,0); std::vector<int> usc(l+1,0);
  if (usa_path) for (int lati=2; lati<=max_paths && lati<l; lati++)
    for (vert[0]=0; vert[0]<l; vert[0]++)
      set_paths(ent, usc, vert, 0, lati, ig, ris, n_spigolo);

	//occhiali 
	if (scrivi_filtri) std::cout << "Occhiali: " << std::endl;
  if (usa_occhiali && max_cicli>=2) for (int i=0; i<l; i++) if (ig.tipo(i)==0) {
		//mette in usc[0] e ent[0] le due uscite che vanno verso lo stesso vertice adiacente
    std::vector<int> usc(2); std::vector<int> ent(2);
		for (int j=0; j<4; j++)
			for (int k=j+1; k<4; k++)
				if (ig.get_v(i,j) == ig.get_v(i,k)) {
					usc[0] = j; ent[0] = k;
				}
		//mette in altre_usc[2] le due altre uscite dal vertice
		std::vector<int> altre_usc(0);
		for (int j=0; j<4; j++)
			if (j!=usc[0] && j!=ent[0])
				altre_usc.push_back(j);
		//considera separatamente le due altre_usc[]
		for (int t=0; t<2; t++) {
			int i2 = ig.get_v(i,altre_usc[t]);
			if (ig.tipo(i2)==0) {
				//controlla che siano in ordine (usc[0], ent[0], altre_usc[t], altre_usc[1-t])
				//se non lo sono permuta usc[0] e ent[0]
				int usc_temp[4];
				usc_temp[0] = usc[0];
				usc_temp[1] = ent[0];
				usc_temp[2] = altre_usc[t];
				usc_temp[3] = altre_usc[1-t];
				if (segno(usc_temp))
					std::swap(usc[0], ent[0]);
				//crea gli altri ent, usc, ent2, usc2
				ent[1] = ig.get_col(i,usc[0]);
				usc[1] = ig.get_col(i,ent[0]);
				std::vector<int> usc2(2); std::vector<int> ent2(2);
				for (int j=0; j<4; j++)
					for (int k=j+1; k<4; k++)
						if (ig.get_v(i2,j) == ig.get_v(i2,k)) {
							usc2[0] = j; ent2[0] = k;
						}
				std::vector<int> altre_usc2(0);
				//mette in altre_usc2[0] la stanghetta centrale degli occhiali
				for (int j=0; j<4; j++)
					if (j!=usc2[0] && j!=ent2[0])
						altre_usc2.push_back(j);
				if (ig.get_v(i2,altre_usc2[0]) != i)
					std::swap(altre_usc2[0], altre_usc2[1]);
				//controlla che siano in ordine (usc2[0], ent2[0], altre_usc2[1], altre_usc2[0])
				//se non lo sono permuta usc2[0] e ent2[0]
				usc_temp[0] = usc2[0];
				usc_temp[1] = ent2[0];
				usc_temp[2] = altre_usc2[1];
				usc_temp[3] = altre_usc2[0];
				if (segno(usc_temp))
					std::swap(usc2[0], ent2[0]);
				//finisce di creare ent2, usc2
				ent2[1] = ig.get_col(i2,usc2[0]);
				usc2[1] = ig.get_col(i2,ent2[0]);
				std::vector<torsione> tor(2,0);
				std::vector<torsione> tor2(2,0);
				std::vector<int> pos(2,0);
				std::vector<int> pos2(2,0);
        //tor[j]+2 e tor2[j]+2 danno i due bigoni embedded
				::ciclo (usc, ent, tor, pos, 2);
				::ciclo (usc2, ent2, tor2, pos2, 2);
				for (int j=0; j<2; j++) {
					tor[j]+=1;
					tor2[j]+=1;
				}
				//modifica ent[1], usc[1], ent2[1], usc2[1] per trovare la torsione della stanghetta
				ent[1] = 1; usc[1] = 0; ent2[1] = 1; usc2[1] = 0;
				std::vector<torsione> tor_temp(2,0);
				::ciclo (usc, ent, tor_temp, pos, 2);
				torsione tor_stanghetta = tor_temp[0]+1;
				::ciclo (usc2, ent2, tor_temp, pos, 2);
				tor_stanghetta += (tor_temp[0]+1);
				//esclude terne
				torsione esclusi[4][3] = {{0,0,0},{0,2,1},{1,1,1},{1,2,0}};
				int spigolo[5];
				//gli spigoli del filtro sono ordinati come a,b,c,d,e
				spigolo[0] = n_spigolo[i][ent[0]];
				spigolo[1] = n_spigolo[i][usc[0]]; 
				spigolo[2] = n_spigolo[i2][usc2[0]]; 
				spigolo[3] = n_spigolo[i2][ent2[0]];
				spigolo[4] = n_spigolo[i][altre_usc[t]];
				for (int h=0; h<=1; h++)
					for (int j=2; j<=3; j++) 
						for (int k=0; k<4; k++) {
							filtro nuovo;
							nuovo.push_back(spigolo_e_tor(spigolo[h], esclusi[k][0] + tor[0]));
							nuovo.push_back(spigolo_e_tor(spigolo[4], esclusi[k][1] + tor_stanghetta));
							nuovo.push_back(spigolo_e_tor(spigolo[j], esclusi[k][2] + tor2[0]));
							aggiungi(ris, nuovo);
						}
			}
		}
	}

	// orbite. Funziona bene solo se ig e' in forma minima.
	// filtra alcuni ografi che non sono in forma minima.
	if (scrivi_filtri) std::cout << "Orbite: " << std::endl;
  if (usa_orbite) {
		std::list<trasformazione> automorfismi = ig.trasformazioni_minimizzanti();

		//elenca in 'orb' gli spigoli che sono l'orbita dello spigolo 0 sotto l'azione degli automorfismi
		//orb[][0] = p(spigolo) per ogni automorfismo p
		//se 'tor' e' la torsione di 'spigolo'
		//allora orb[][2]+tor e' la torsione di p(spigolo) se non c'e' specchio
		//et     orb[][2]-tor se c'e' specchio
		tensor2 orb;
		for (std::list<trasformazione>::const_iterator p=automorfismi.begin(); p!=automorfismi.end(); p++) {
			ografo o(ig, std::vector<torsione>(2*l,0));
			o.trasforma(*p);
			int vert_dopo = p->t[0][0];
			int usc_dopo = 0;
			if (p->t[1][0]) usc_dopo = RR[0];
			int mossa = p->t[2][0];
			usc_dopo = Aquattro[mossa][usc_dopo];
			std::vector<int> nuovo_spigolo(3,0);
			nuovo_spigolo[0] = n_spigolo[vert_dopo][usc_dopo];
			nuovo_spigolo[1] = p->t[1][0];
			nuovo_spigolo[2] = o.get_tor(vert_dopo, usc_dopo).get_t();
			orb.push_back(nuovo_spigolo);
		}
		elimina_duplicati(orb);

		if (scrivi_filtri) { 
			for (int i=0; i<orb.size(); i++)
				std::cout << "(" << orb[i][0] << "," << orb[i][1] << "," << orb[i][2] << ")";
			std::cout << std::endl;
		}
		for (int t=1; t<orb.size(); t++)
			if (orb[t][0]!=orb[0][0])
				for (int tor1=0; tor1<4; tor1++)
					for (int tor2=0; tor2<tor1; tor2++) {
						torsione tor2_immagine = tor2;
						if (orb[t][1]) tor2_immagine.invert();
						tor2_immagine += orb[t][2];
						filtro nuovo;
						nuovo.push_back(spigolo_e_tor(orb[0][0], tor1));
						nuovo.push_back(spigolo_e_tor(orb[t][0], tor2_immagine));
						aggiungi(ris, nuovo);
					}
			else for (int tor=0; tor<3; tor++) {
				torsione tor_immagine = tor;
				if (orb[t][1]) tor_immagine.invert();
				tor_immagine += orb[t][2];
				if (tor_immagine.get_t() < tor) {
					filtro nuovo(1, spigolo_e_tor(orb[0][0], tor));
					aggiungi(ris, nuovo);
				}
			}
	}
	return ris;
}

void set_paths(std::vector<int>& ent, std::vector<int>& usc, std::vector<int>& vert, int liv, int lati, 
							 const igrafo& ig, std::vector<std::vector<std::list<filtro> > >& ris,
							 const tensor2& n_spigolo) {
  if (liv<lati) {
    for (usc[liv]=0; usc[liv]<4; usc[liv]++) if (!liv || usc[liv]!=ent[liv]) {
      ent[liv+1] = ig.get_col(vert[liv], usc[liv]);
      vert[liv+1] = ig.get_v(vert[liv], usc[liv]);
			//vado avanti solo se il nuovo vertice non compare gia':
      if (!std::count (&vert[0], &vert[liv+1], vert[liv+1]))
				set_paths(ent, usc, vert, liv+1, lati, ig, ris, n_spigolo);
    }
  }
	//Evitiamo di considerare lo stesso path due volte;
	//se lati==2, proseguiamo pero' comunque perche' sotto c'e' il Filtro 4.
  else if (vert[0] < vert[lati] || lati == 2) {
    std::vector<torsione> tor(lati,0); std::vector<int> pos(lati,0); ent[0]=0; //va bene qualsiasi cifra 0,1,2
    ::ciclo (usc, ent, tor, pos, lati);			                         //coordinate delle torsioni 2
    std::vector<int> spigolo(lati,0); 
    for (int i=0; i<lati; i++)
			spigolo[i] = n_spigolo[vert[i]][usc[i]];
		//evita di considerare lo stesso path due volte. (Necessario chiedere ancora nel caso lati == 2)
		if (vert[0] < vert[lati]) {
			filtro nuovo;
			for (int i=0; i<lati; i++)
				nuovo.push_back(spigolo_e_tor(spigolo[i], (i==0 || i==lati-1) ? 3:tor[i]));
			aggiungi (ris, nuovo);
		}
		
		//Filtro 4 (bigono e path):
		if ((lati == 2) && ig.tipo(vert[2])==0) {
			//mette in usc[2], ent[4] le due uscite del bigono
			for (int t=0; t<4; t++)
				for (int u=t+1; u<4; u++)
					if (ig.get_v(vert[2],t) == ig.get_v(vert[2],u)) {
						usc[2] = t; ent[4] = u;
					}
			vert[4] = ig.get_v(vert[2],usc[2]);
			//controlla che v non sia un vertice gia' visto
			//e mette in usc[4] la quarta uscita di vert[2]
			if (std::count (&vert[0], &vert[3], vert[4]) == 0) {
				for (int t=0; t<4; t++)
					if (t!=ent[2] && t!=usc[2] && t!=ent[4])
						usc[4] = t;
				//controlla che (usc[4], ent[4], usc[2], ent[2]) siano ordinate,
				//altrimenti permuta usc[2] e ent[4]:
				int usc_temp[4];
				usc_temp[0] = usc[4];
				usc_temp[1] = ent[4];
				usc_temp[2] = usc[2];
				usc_temp[3] = ent[2];
				if (segno(usc_temp))
					std::swap(usc[2], ent[4]);
				//costruisce i mancanti ent[3], usc[3], spigolo[2], spigolo[3]:
				ent[3] = ig.get_col(vert[2], usc[2]);
				usc[3] = ig.get_col(vert[2], ent[4]);
				spigolo.resize(4); tor.resize(5); pos.resize(5);
				for (int t=2; t<=3; t++)
					spigolo[t] = n_spigolo[vert[t]][usc[t]];
				//la curva che percorre tutto il path e' data dalle torsioni tor[i] + 2:
				::ciclo(usc, ent, tor, pos, 5);
				for (int i=0; i<4; i++)
					tor[i]+=1;
				//Filtro 4, nelle coordinate usate qui:
				torsione esclusi[2][3] = {{3,1,2},{3,2,1}};
				for (int i=0; i<=1; i++) {
					filtro nuovo(0);
					for (int j=0; j<=2; j++)
						nuovo.push_back(spigolo_e_tor(spigolo[j], tor[j] + esclusi[i][j]));
					aggiungi(ris, nuovo);
				}
			}
		}
  }
}

void set_ciclo(std::vector<int>& ent, std::vector<int>& usc, const std::vector<int>& vert, int liv, int lati, 
								 const igrafo& ig, std::vector<std::vector<std::list<filtro> > >& ris,
								 const tensor2& n_spigolo) {
	if (liv<lati) {
    for (usc[liv]=0; usc[liv]<4; usc[liv]++)
      if (vert[(liv+1) % lati] == ig.get_v(vert[liv], usc[liv]))
				if (lati > 2 || liv==0 || ig.get_col(vert[1], usc[1])> usc[0])
					set_ciclo(ent, usc, vert, liv+1, lati, ig, ris, n_spigolo);
	}
  else {
    for (int i=0; i<lati; i++) ent[(i+1) % lati] = ig.get_col(vert[i], usc[i]);
    std::vector<torsione> tor(lati,0); std::vector<int> pos(lati,0);
    ::ciclo (usc, ent, tor, pos, lati);
    for (int i=0; i<lati; i++) tor[i]+=1;
		//a questo punto, tor[0]+2,...,tor[lati-1]+2 danno il lati-gono embedded
		//abbiamo quindi le stesse coordinate usate in Filtro 2
		std::vector<int> spigolo(lati,0);
		for (int t=0; t<lati; t++)
			spigolo[t] = n_spigolo[vert[t]][usc[t]];
		
		//Usiamo il Filtro 2, facendo variare il vertice di partenza:
		for (int primo = 0; primo < lati; primo++) {

			//Filtro 2.2: (x, 2, ..., 2)  per x=0,1  
			for (int x = 0; x<=1; x++) {
				filtro nuovo(1, spigolo_e_tor(spigolo[0], tor[0] + x));
				for (int t=1; t<lati; t++)
					nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t]+2));
				aggiungi(ris, nuovo);
			}
			
			//Filtro 2.3:                  (0, 2, ..., 2, 1, 2, ..., 2)
			//Filtro 2.2 + 2.4 (rel a 2.1):(x, 2, ..., 2, 3, 2, ..., 2) per x = 0,1
			torsione y[3][2] = {{0,1},{0,3},{1,3}};
			for (int i=0; i<3; i++) { 
				for (int posto = 1; posto<lati; posto++) {
					filtro nuovo(1, spigolo_e_tor(spigolo[0], tor[0] + y[i][0]));
					for (int t=1; t<posto; t++)
						nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t]+2));
					nuovo.push_back(spigolo_e_tor(spigolo[posto], tor[posto] + y[i][1]));
					for (int t=posto+1; t<lati; t++)
						nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t]+2));
					aggiungi(ris, nuovo);
				}
			}

			//Filtro 2.5 + 2.4 (rel a 2.3 caso n=4):
			torsione z[5][4] = {{1,0,1,0},{1,0,1,3},{0,1,0,3},{0,2,1,3},{0,3,1,2}};
			if (lati==4) {
				for (int i=0; i<5; i++) {
					filtro nuovo;
					for (int t=0; t<4; t++)
						nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t] + z[i][t]));
					aggiungi(ris, nuovo);
				}
			}
			
			//Filtro 2.6 e 2.4 (rel a 2.3 caso n=5), ho messo l>7: 
			torsione w[][5] = {{1,0,1,0,2},{0,1,0,1,2},{1,0,1,2,3},{0,1,0,2,3},
												 {1,0,1,0,3},{0,1,0,1,3},{1,0,1,3,2},{0,1,0,3,2},
												 {0,2,1,3,2},{0,2,1,2,3},
												 {1,2,0,3,2},{1,2,0,2,3}};
			if (lati == 5 && ig.lung()>7) for (int i=0; i<12; i++) {
				filtro nuovo;
				for (int t=0; t<5; t++)
					nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t] + w[i][t]));
				aggiungi(ris, nuovo);
			}

			std::rotate(spigolo.begin(), spigolo.begin() + 1, spigolo.end());
			std::rotate(tor.begin(), tor.begin() + 1, tor.end());
		}

		//Quadrato: vede se con disc replacement si ottengono igrafi gia' visti
		//rallenta molto con l<=7, ho messo l>8 
		if (lati == 4 && usa_quadrato && ig.lung()>8) {
			std::vector<torsione> tors(2*ig.lung(),0);
			for (int i=0; i<4; i++)
				tors[spigolo[i]] = tor[i] + 2;
			//crea un ografo con un quadrato su spigolo[0], ..., spigolo[3];
			ografo o(ig, tors);
			//cerca la faccia corrispondente in o:
			int num_faccia = -1;
			bool trovata=FALSE;
			while (!trovata) {
				if (num_faccia+1 >= o.get_ff().size())
					int tre = 4;
				faccia f = o.get_ff()[++num_faccia];
				if (f.lati()==4 && f.embedded()) {
					bool ok = 1;
					for (int t=0; t<4; t++) {
						int spig = n_spigolo[f.get_passaggio(t)][f.get_uscita(t)];
						if (std::count(&spigolo[0], &spigolo[4], spig)!=1)
							ok=0;
					}
					if (ok) trovata = 1;
				}
			}
			//fa un disc replacement (ce ne sono due tipi):
			for (int tipo=0; tipo<=1; tipo++) {
				ografo o2 = o;
				o2.disc_replacement(num_faccia, (bool) tipo);  //e' importante che sia bool!
				//se l'igrafo soggiacente a o2 e' < ig, crea filtro:
				o2.set_forma_minima();
				if (((igrafo) o2) < ig) { //ho tolto l'utile
					filtro nuovo;
					for (int t=0; t<4; t++)
						nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t] + 2));
					aggiungi(ris, nuovo);
					break;
				}
			}
		}
		
		//Filtro 2.1
		if (lati<=3) {
			filtro nuovo;
			for (int t=0; t<lati; t++)
				nuovo.push_back(spigolo_e_tor(spigolo[t], tor[t] + 2));
			aggiungi(ris, nuovo);
		}
	
		//Filtro 1 nel caso n==2 (che set_path non considera se il path e' chiuso!)
		if (lati==2) {
			filtro nuovo;
			for (int t=0; t<2; t++)
				nuovo.push_back(spigolo_e_tor(spigolo[t], 3));
			aggiungi(ris, nuovo);
		}

		//Filtri 3 e 8
    if (lati==3 && usa_scorpione) {
			if (scrivi_filtri) std::cout << "Filtri 3 e 8 (scorpione) :" << std::endl;
 			//t e' il vertice del triangolo:
      for (int t=0; t<=2; t++) {
				//mette in altra_usc[2] le due uscite che non si conoscono ancora:
				int altra_usc[2]; int tt=0;
				for (int j=0; j<4; j++) 
					if (j!=usc[t] && j!=ent[t])
						altra_usc[tt++]=j;
				//controlla che l'insieme ordinato di uscite (altra_usc[0],altra_usc[1],usc[t],ent[t])
				//(che corrispondera' sotto a (spigolo[3], spigolo[4], spigolo[0], spigolo[2]))
				//possa essere ottenute con mosse C, altrimenti permuta altra_usc[]
				int usc_temp[4];
				for (int i=0; i<2; i++) usc_temp[i]=altra_usc[i];
				usc_temp[2]=usc[t]; usc_temp[3]=ent[t];
				if (segno(usc_temp))
					std::swap(altra_usc[0], altra_usc[1]);
				//mette in vicino[2] i due vicini nuovi
				int vicino[2];
				for (int i=0; i<2; i++) vicino[i]=ig.get_v(vert[t], altra_usc[i]);
				//(spigolo[0], spigolo[1], spigolo[2]) e' il triangolo (vert[t] e' tra spigolo[2] e spigolo[0])
				//e spigolo[3], spigolo[4] sono i due nuovi:
				//          |           |
				//          |4          |
				//       3  |     0     |
				//     -----------------|-----
				//          |           |
				//          \           |1
				//           \  2       |
				//            ----------------
				//                      |
				spigolo.resize(5);
				for (int i=0; i<=1; i++) spigolo[3+i] = n_spigolo[vert[t]][altra_usc[i]];

				//Filtro 3: nei casi 0 e 1 si considerano gli spigoli 3 e 4
				for (int caso=0; caso<=1; caso++) {
					//controlla che non sia contenuto in un cappello:
					if (!std::count(&vert[0], &vert[3], ig.get_v(vert[t], altra_usc[caso]))) {
						filtro nuovo(1, spigolo_e_tor(spigolo[3],3)); 
						nuovo.push_back(spigolo_e_tor(spigolo[1],tor[1] + 2));
						aggiungi(ris, nuovo);
						torsione esclusi[3][2] = {{0,2},{1,0},{2,1}};
						for (int i=0; i<3; i++) {
							nuovo = filtro(1, spigolo_e_tor(spigolo[3],3)); 
							nuovo.push_back(spigolo_e_tor(spigolo[0],tor[0] + esclusi[i][0]));
							nuovo.push_back(spigolo_e_tor(spigolo[2],tor[2] + esclusi[i][1]));
							aggiungi(ris, nuovo);
						}
						for (int i=0; i<=1; i++) {
							filtro nuovo(1, spigolo_e_tor(spigolo[3],3)); 
							for (int j=0; j<=2; j++)
								nuovo.push_back(spigolo_e_tor(spigolo[j],tor[j] + i));
							aggiungi(ris, nuovo);
						}
					}
					//scambia (0 con 2) e (3 con 4) -- dopo la seconda volta tornano come prima.
					std::swap(spigolo[3], spigolo[4]); 
					std::swap(spigolo[0], spigolo[2]);	std::swap(tor[0], tor[2]);
				}

				//Filtro 8:
				if (vicino[0]==vicino[1]) {
					//mette in tor2[0], tor2[1] le torsioni che danno il bigono:
					std::vector<int> ent2(2,0); std::vector<int> usc2(2,0); std::vector<torsione> tor2(2,0); std::vector<int> pos2(2,0);
					usc2[0]=altra_usc[0]; ent2[0]=altra_usc[1];
					usc2[1]=ig.get_col(vert[t], ent2[0]);
					ent2[1]=ig.get_col(vert[t], usc2[0]);
					::ciclo (usc2, ent2, tor2, pos2, 2);
					//fa in modo che 2+torsione[3], 2+torsione[4] siano le torsioni che danno il bigono:
					tor.resize(5);
					for (int i=0; i<=1; i++) tor[3+i] = tor2[i]+1;
					//triplette:
					torsione esclusi[8][3] = {{1,2,1},{1,0,2},{1,1,0},{1,0,3},{0,2,0},{0,0,1},{0,1,2},{0,1,3}};
					//non c'e' bisogno di scrivere le equazioni per spigolo[3] e spigolo[4]. Ne sceglie uno:
					int primo = (spigolo[3] < spigolo[4]) ? 3:4;
					for (int j=0; j<=2; j+=2)
						for (int k=0; k<8; k++) {
							filtro nuovo;
							nuovo.push_back(spigolo_e_tor(spigolo[primo], tor[primo] + esclusi[k][0]));
							nuovo.push_back(spigolo_e_tor(spigolo[j], tor[j] + esclusi[k][1]));								
							nuovo.push_back(spigolo_e_tor(spigolo[1], tor[1] + esclusi[k][2]));								
							aggiungi(ris, nuovo);
						}
					//quadruplette:
					for (int k=0; k<=1; k++) {
						filtro nuovo;
						nuovo.push_back(spigolo_e_tor(spigolo[primo], tor[primo] + k));
						for (int h=0; h<=2; h++)
							nuovo.push_back(spigolo_e_tor(spigolo[h], tor[h] + k));
						aggiungi(ris, nuovo);
					}
				}
				spigolo.resize(3);
				tor.resize(3);
				std::rotate(spigolo.begin(), spigolo.begin() + 1, spigolo.begin() + 3); 
				std::rotate(tor.begin(), tor.begin() + 1, tor.begin() + 3);    
			}
			if (scrivi_filtri) std::cout << "Ancora cicli :" << std::endl;
		}
		
		//Casa: 
		if (usa_casa && lati == 4) {
			//vertice t-esimo:
			for (int t=0; t<4; t++) {
				//guarda se ci sono due spigoli tra vert[t] e vert[t+1]:
				int quanti = 0;
				for (int j=0; j<4; j++)
					if (ig.get_v(vert[t],j) == vert[(t+1) % 4])
						quanti++;
				if (quanti == 2) {
					//mette le altre uscite in altra_usc[2][2]:
					int altra_usc[2][2];
					for (int i=0; i<=1; i++) {
						int tt=0;
						for (int j=0; j<4; j++) 
							if (j!=usc[t] && j!=ent[t])
								altra_usc[i][tt++]=j;
					}
					//controlla che l'insieme ordinato (altra_usc[0][0],altra_usc[0][1],usc[t],ent[t])
					//possa essere ottenuto con mosse C, altrimenti permuta altra_usc[0][],
					//ed idem per (altra_usc[1][0], altra_usc[1][1], usc[t+1], ent[t+1]):
					int usc_temp[4];
					for (int j=0; j<=1; j++) {
						for (int i=0; i<=1; i++) usc_temp[i]=altra_usc[j][i];
						usc_temp[2]=usc[(t+j) % 4]; usc_temp[3]=ent[(t+j) % 4];
						if (segno(usc_temp))
							std::swap(altra_usc[j][0], altra_usc[j][1]);
					}
					//guarda se i due spigoli sono incidenti nel modo giusto:					
					for (int i=0; i<=1; i++)
						if (ig.get_v(vert[t], altra_usc[0][i]) == vert[(t+1) % 4] && 
								ig.get_v(vert[(t+1) % 4], altra_usc[1][i]) == vert[t]) {
							filtro nuovo(1, spigolo_e_tor(spigolo[(t+2) % 4], tor[(t+2) % 4] + 2));
							aggiungi(ris, nuovo);
						}
				}	
			}
		}
	}
}

//dice se ci sono (e quali) filtri che escludono l'ografo 'o'
void test_restrizioni(const ografo& o) {
	std::vector<std::vector<std::list<filtro> > > filtri = filtri_brick(o);
	const int l = o.lung();
	std::vector<int> tor(0);
	for (int i=0; i<l; i++)
		for (int j=0; j<4; j++)
			if (o.prima_occorrenza_spigolo(i,j))
				tor.push_back(o.get_tor(i,j).get_t());
	std::cout << "Filtri che toccano l'ografo : " << std::endl;
	for (int spigolo = 0; spigolo < 2*l; spigolo++)
		for (f_iterator fi=filtri[spigolo][tor[spigolo]].begin(); fi!=filtri[spigolo][tor[spigolo]].end(); fi++) {
			st_iterator sti = fi->begin();
			while (sti!=fi->end() && tor[sti->spigolo]==sti->tor) sti++;
			if (sti == fi->end()) {
				std::cout << "(" << spigolo << "," << tor[spigolo] << ") ";
				scrivi(*fi);
			}
		}
}

//scrive i filtri
void scrivi(const std::vector<std::vector<std::list<filtro> > >& ris) {
	for (int i=0; i<ris.size(); i++)
		for (int j=0; j<ris.size(); j++)
			for (f_c_iterator fi = ris[i][j].begin(); fi!=ris[i][j].end(); fi++) {
				std::cout << "(" << i << "," << j << ") ";
				scrivi(*fi);
			}
}

void scrivi(const filtro& f) {
	for (st_c_iterator sti = f.begin(); sti!=f.end(); sti++)
		std::cout << "(" << sti->spigolo << "," << sti->tor.get_t() << ") ";
	std::cout << std::endl;
}

torsione maschera (int usc, int col, int usc2, int col2) {
  torsione tor=0;
  if (mod(usc,2)) {usc=3-usc; usc2=3-usc2;}
  if (usc!=0) {usc=0; usc2=mod(usc2+2,4);}
  if (mod(col,2)) {col=3-col; col2=3-col2;}
  if (col!=0) {col=0; col2=mod(col2+2,4);}
  while (usc2!=3) {
    ++tor;
		usc2++;
  }
  while (col2!=1) {
    ++tor;
		col2 = (col2==3) ? 1 : col2+1;
  }
  return tor;
}

//se c'e' un ciclo in cui ogni vertice ha spigoli nelle 2 direzioni usc[i], ent[i],
//mette in tor[i] la torsione dell'n-gono embedded
void ciclo (const std::vector<int>& usc, const std::vector<int>& ent, std::vector<torsione>& tor, 
						std::vector<int>& pos, int nv) {
	int l = usc.size();
  std::vector<int> u(l,0);
  std::vector<int> e(l,0);
  std::vector<int> p(l,0);
  for (int t=0; t<nv; t++) {
    tor[t]=2; u[t]=usc[t]; e[t]=ent[t];
    pos[t] = mod(usc[t]+1,4);
    if (pos[t]==ent[t]) pos[t]=mod(usc[t]+2,4);
    p[t]=pos[t];
  }
  for (int t=0; t<nv; t++) {
    if (mod(u[t],2)) {u[t]=3-u[t]; e[t]=3-e[t]; p[t]=3-p[t];}
    if (u[t]!=0) {u[t]=0; e[t]=mod(e[t]+2,4); p[t]=mod(p[t]+2,4);}
  }
  for (int t=0; t<nv; t++)
    while (e[t]!=1) {
      int t2 = mod(t-1+nv,nv);
      ++tor[t]; ++tor[t2];
      e[t] = e[t]==3 ? 1 : e[t]+1;
      p[t] = p[t]==3 ? 1 : p[t]+1;
    }
  for (int t=0; t<nv; t++)
    if (p[t]==2) {
      bool trovato = FALSE;
      for (int t1=0; t1<4 && !trovato; t1++)
				if (t1!=usc[t] && t1!=ent[t] && t1!=pos[t]) {
					trovato = TRUE;
					pos[t]=t1;
				}
    }
}

torsione Cforza (int* c1, int* c2, int* c3, int* c4, int cf1, int cf2) {
  if (mod(abs(*c1-cf1),2)) {
    *c1=3-*c1; if (c2) *c2=3-*c2; if (c3) *c3=3-*c3; if (c4) *c4=3-*c4;
  }
  if (*c1!=cf1) {
    *c1=cf1; if (c2) *c2=mod (*c2+2,4); if (c3) *c3=mod (*c3+2,4); if (c4) *c4=mod (*c4+2,4);
  }
  bool parita = mod(*c1,2);
  int tor=0;
  if (c2) while (*c2!=cf2) {
    *c2 = (mod (*c2+1,4) == *c1) ? mod(*c2+2,4):mod(*c2+1,4);
    if (c3) *c3 = (mod (*c3+1,4) == *c1) ? mod(*c3+2,4):mod(*c3+1,4);
    if (c4) *c4 = (mod (*c4+1,4) == *c1) ? mod(*c4+2,4):mod(*c4+1,4);
    tor+=(1+parita);
  }
  return tor;
}

void cappello (int j1, int c1, int j2, int c2, int k, int ck, int h, int ch, 
							 std::vector<torsione>& tor, std::vector<int>& c) {
  for (int t=0; t<4; t++) tor[t]=0;
  torsione t = Cforza(&j1, &j2, &k, 0, 0, 1); c[0]=k;
  tor[0]=t; tor[1]=t; tor[2]=t;
  t = Cforza(&c1, &c2, &h, 0, 1, 0); c[1]=h;
  tor[0]+=t; tor[1]+=t; tor[3]=t;
  t = Cforza(&ck, &ch, 0, 0, 1, 2);
  tor[2]+=t; tor[3]+=t;
}

