#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <list>
#include <vector>
#include <algorithm>

#include "defs.h"
#include "math.h"
#include "templates.cpp"
#include "string.h"
#include "graph.h"

bool mostra_contatore = FALSE;

//costruttore standard
vertice::vertice() {for(int t=0; t<4; t++) vicino[t]=-1;}

//costruttore da stringa del tipo "3471"
vertice::vertice(const int* w) {for(int t=0; t<4; t++) vicino[t]=w[t];}

//distruttore
grafo::~grafo() {
	disalloca(v);
}

//costruttore di copia
grafo::grafo (const grafo& g) {
	for (int i=0; i<g.lung(); i++)
		v.push_back(new vertice(*g[i]));
}

//costruttore da matrice w[][4]={{a,b,c,d},{e,f,g,h},...} di dimensione l*4
grafo::grafo (const tensor2& w, int l) {
  for (int i=0; i<l; i++)
    v.push_back(new vertice(&w[i][0]));
}

//carica (dal file "schel#.dat" con #=l) il grafo numero ind (partendo da 1) con l vertici
grafo::grafo (int l, long ind) {
  char z[14] = "schel"; strcat(z,l); strcat (z, ".dat");
  std::ifstream f(z, std::ios::in);
  long passo = 27+18*l;
  f.seekg(passo*(ind-1), std::ios::beg);
  *this = grafo (f);
}

//carica grafo dal file 'file'
grafo::grafo (std::istream& file, int tipo_formato) {
  char w[100];
  file.getline(w, 100);
  const int l=ctoi(&w[11],2);
  for (int i=0; i<l; i++) {      
    file.getline(w,100);
    int nuovo[4];
    for (int j=0; j<4; j++)
      nuovo[j] = ctoi(&w[9+2*j],2);
    v.push_back(new vertice(nuovo));
  }
  file.getline(w,100);  //prende la linea di chiusura '--------'
}

//aggiunge num vertici: da usare solo se ce ne e' almeno uno!
//infatti copia la struttura di questo (che puo' essere derivata da vertice)
void grafo::aggiungi_vertici(int num) {
	for (int i=0; i<num; i++)
		v.push_back(v[0]->clone());
}

//permuta il grafo secondo 'p'. Esempio: se p = {2,1,3,0}, il vecchio v[0]
//viene spostato in posizione numero 2, quindi corrisponde al nuovo v[2]
void grafo::trasforma (const trasformazione& trasf) {
  const std::vector<int> p = trasf.t[0];
  const int l=lung();
	tensor2 old_v(l,4);
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      old_v[i][j] = get_v(i,j);
  ::permuta(v, p);
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++)
      v[p[i]]->set_vicino(j,p[old_v[i][j]]);
}

//dice di che tipo e' il vertice j-esimo
int grafo::tipo(int j) const {
  return v[j]->tipo(j);
}

//dice se il grafo, senza gli spigoli che soddisfano w(vert,col)==1, e' connesso.
bool grafo::connesso(const tensor2& w) const{
  const int l=lung(); int contatore=0;
  std::vector<int> cc(l,0);
  for (int t=0; t<l; t++) {
    if (!cc[t]) cc[t] = ++contatore;
    for (int r=0; r<4; r++) {
      const int colore = cc[get_v(t,r)];
      if (!w[t][r] && colore != cc[t]) {
        if (!colore) cc[get_v(t,r)] = cc[t];
        else for (int i=0; i<l; i++) {
          if (cc[i]==colore) cc[i] = cc[t];
        }
      }
    }
  }
  contatore=cc[0]; bool connesso=1;
  for (int t=0; connesso && t<l; t++) if (cc[t]!=contatore) connesso=0;
  return connesso;
}

bool grafo::riccio(int i) const {
  return (tipo(i)>=3);
}

int grafo::quanti_spigoli(int vert1, int vert2) const {
	int ris=0;
	for (int t=0; t<4; t++)
		if (get_v(vert1,t) == vert2)
			ris++;
	return ris;
}

//Restituisce 0,1 o 2 vettori. 
//Questi contengono i vertici che compongono una catena, che parte da vert, se esiste.
//(ce ne sono al piu' due)
//si ferma al vertice max, se max>0.
tensor2 catena(const grafo& g, int vert, int max) {
  tensor2 cat(0);
  std::vector<int> gia_visti(1,vert);
  for (int t=0; t<4; t++) {
		int vert2 = g.get_v(vert,t);
		if (!count(gia_visti.begin(), gia_visti.end(), vert2)
				&& g.quanti_spigoli(vert, vert2)>1) {
			std::vector<int> nuova(1,vert); nuova.push_back(vert2);
			gia_visti.push_back(vert2);
			
			bool finito = FALSE;
			while (g.tipo(vert2) == 1 && vert2!=vert) {
				int liv = nuova.size()-1;
				for (int j=0; j<4; j++)
					if (g.get_v(nuova[liv],j) != nuova[liv-1]) 
						vert2 = g.get_v(nuova[liv],j);
				if (vert2!=vert) nuova.push_back(vert2);
			}
			
			if (max>0 && nuova.size()>max) nuova.erase(nuova.begin() + max, nuova.end());
			cat.push_back(nuova);
		}
	}
  return cat;
}

//ricorsiva, serve a cicli(int max)
void cicli(const grafo& g, int liv, int q, std::vector<int>& vert,
														 tensor3& cic) {
  if (!liv) {
    const int l = g.lung();
    for (int j=0; j<l; j++) {
      vert[0]=j;
      cicli(g, 1, q, vert, cic);
    }
  }
  if (liv>0 && liv<q) {
    const int k=vert[liv-1];
    for (int i=0; i<4; i++) 
			if ((i==0 || g.get_v(k,i-1) != g.get_v(k,i)) && g.get_v(k,i)>vert[0]) {
				bool trovato=FALSE;					//vert[0] deve essere il piu piccolo
				for (int j=0; j<liv; j++)
					if (g.get_v(k,i)==vert[j]) trovato=TRUE;
				if (!trovato) {
					vert[liv] = g.get_v(k,i);
					cicli(g, liv+1, q, vert, cic);
				}
			}
  }
  if (liv==q && (q<=2 || vert[liv-1]>vert[1])) { //se q>2 evita di scrivere stesso ciclo 2 volte
    bool trovato=FALSE;
    for (int i=0; i<4 && !trovato; i++) if (g.get_v(vert[liv-1],i) == vert[0]) {
      cic[q].push_back(vert);
      trovato = TRUE;
    }
  }
}

//produce tutti i cicli con lati<=max
//cicli[lt][k] e' quindi un vettore, che rappresenta il k-esimo (partendo da 0) ciclo
//di vertici (etichettati da 0 a lung()-1) avente 'lt' lati
//(quindi cicli[0] e' vuoto)
//Questo vettore e' ordinato (ciclicamente) in modo tale che l'elemento piu' alto sia l'ultimo.
tensor3 cicli(const grafo& g, int max) {
  const int l=g.lung(); if (max>l) max=l;
  tensor3 cic(max+1);
  for (int j=0; j<l; j++) {
    if (g.riccio(j)) cic[1].push_back(std::vector<int>(1,j));
    if (max>1) {
			tensor2 cat = catena(g,j,2);
			for (int t = 0; t<cat.size(); t++)
				if (cat[t][1] > j)
					cic[2].push_back(cat[t]);
		}
  }
  for (int q=3; q<=max; q++) {
    std::vector<int> vert(q);
    cicli (g, 0, q, vert, cic);
  }
  return cic;
}

//funzione ricorsiva che serve a utile()
void grafo::utile(tensor2& w, int i[][4], int j[][4], int liv, int l, bool& utl) const {
  for (i[0][liv] = (liv==0) ? 0:i[0][liv-1]; i[0][liv]<l; i[0][liv]++)
    for (j[0][liv] = (liv && i[0][liv-1]==i[0][liv]) ? j[0][liv-1]+1:0 ; j[0][liv]<4; j[0][liv]++) {
      i[1][liv] = get_v(i[0][liv],j[0][liv]);
      if (i[1][liv] > i[0][liv]) {
        for (j[1][liv]=0; get_v(i[1][liv],j[1][liv]) < i[0][liv]; j[1][liv]++) {}
        if (j[0][liv]>0 && get_v(i[0][liv],j[0][liv]-1) == i[1][liv]) j[1][liv]++;
        for (int t=0; t<2; t++) w[i[t][liv]][j[t][liv]]=1;
        if (liv<3)
          utile (w,i,j,liv+1,l,utl);
        else if (!connesso(w)) {
          bool ammissibile = FALSE;
          std::vector<int> spigoli_toccati(l,0);
          for (int t=0; t<2; t++) for (int t2=0; t2<4; t2++) spigoli_toccati[i[t][t2]]++;
          for (int n=0; n<l; n++) {
            if (spigoli_toccati[n]==4) ammissibile = TRUE;
            if (spigoli_toccati[n]==2) {
              const int tp = tipo(n);
              if (!ammissibile && (tp==1 || tp==0)) { 		//cerca catena
                tensor2 cat = catena(*this,n);
                cat = catena(*this,cat[0].back());
                for (int t=0; t<cat[0].size(); t++)
                  if (cat[0][t]!=n && spigoli_toccati[cat[0][t]]==2) {
                    bool sono_quattro_spigoli_distinti = TRUE;
                    for (int t2=0; t2<4; t2++) 
                      if (i[0][t2]!=cat[0][t] && i[1][t2]!=cat[0][t] && i[0][t2]!=n && i[1][t2]!=n) 
                        sono_quattro_spigoli_distinti = FALSE;
                    if (sono_quattro_spigoli_distinti) ammissibile = TRUE;
                  }
              }
              if (!ammissibile && (tp==-1 || tp==0)) {              //cerca "triangolo"
                int vicini[2]; int trovati=0;
                for (int t=0; t<4; t++)
                  for (int t2=0; t2<2; t2++)
                    for (int t3=0; t3<4; t3++)
                      if (spigoli_toccati[get_v(n,t)]==1 && get_v(n,t)==i[t2][t3] && n!=i[1-t2][t3])
                        vicini[trovati++] = get_v(n,t);
                if (trovati == 2)
                  for (int t=0; t<4; t++)
                    if (get_v(vicini[0],t) == vicini[1]) ammissibile = TRUE;
              }
            }
          }
          if (!ammissibile)
            utl = FALSE;
        }
        for (int t=0; t<2; t++) w[i[t][liv]][j[t][liv]]=0;
      }
    }
}

//dice se il grafo e' utile
bool grafo::utile() const {
  if (!irriducibile()) return FALSE;
  bool tutti_tipo1 = TRUE;
  const int l=lung();
  for (int t=0; t<l; t++) if (tipo(t)!=1) tutti_tipo1 = FALSE;
  if (tutti_tipo1 || l==1) return TRUE;
  tensor2 w(l,std::vector<int>(4,0));
  int i[2][4]; int j[2][4];
  bool utl = TRUE;
  utile (w, i, j, 0, l, utl);
  return utl;
}

//dice se e' irriducibile, ovvero se non contiene 1-ponti
bool grafo::irriducibile() const {
  const int l=lung();
  for (int i=0; i<l; i++) if (tipo(i)>=2) return FALSE;
  tensor2 w(l,std::vector<int>(4,0));
	for (int i=0; i<l; i++) for (int j=0; j<4; j++) w[i][j]=0;
  for (int i=0; i<l; i++)
    for (int j=0; j<4; j++) if (get_v(i,j)>i) {
      const int k = get_v(i,j);
			int numero_spigoli = 0;
			for (int t=0; t<4; t++) if (get_v(i,t) == k) numero_spigoli++;
      if (numero_spigoli == 2) {
				for (int t=0; t<4; t++) if (get_v(i,t) == k) w[i][t]=1;
        for (int t=0; t<4; t++) if (get_v(k,t) == i) w[k][t]=1;
        if (!connesso(w))
          return FALSE;
        for (int t=0; t<4; t++) {w[i][t]=0; w[k][t]=0;}
      }
      if (numero_spigoli == 1) {
        int col=-1; for (int t=0; t<4; t++) if (get_v(k,t)==i) col=t;
        w[i][j]=1; w[k][col]=1;
        for (int i2=i; i2<l; i2++)
          for (int j2 = (i==i2) ? j+1 : 0; j2<4; j2++) if (get_v(i2,j2)>i2) {
            const int k2 = get_v(i2,j2);
            int col2=-1; for (int t=0; t<4; t++) if (get_v(k2,t)==i2) col2=t;
            w[i2][j2]=1; w[k2][col2]=1;
            if (!connesso(w))
              return FALSE;
            w[i2][j2]=0; w[k2][col2]=0;
          }
        w[i][j]=0; w[k][col]=0;
      }
    }
  return TRUE;
}

//scrive il grafo come stringa su v
void grafo::scrivi(std::ostream& f, int tipo_formato) const {
  const int l=lung();
  f << "N.Vertici: " << l << "\n";
  for (int i=0; i<l; i++) {
    f << "N. " << i;
		if (i<10) f << " ";
		f << ":   ";
		for (int j=0; j<4; j++)
			f << std::setw(2) << get_v(i,j);
    f << "\n";
  }
}

//salva il grafo sul file f
void grafo::salva(std::ostream& f, int tipo_formato) const {
	scrivi(f, tipo_formato);
	f << "------------\n";
}

//restituisce la parita' del numero di trasposizioni necessarie per ordinare gli elementi del vertice
//30.01.02: trasportata la funzione in ogmath.segno(int* )
bool vertice::segno() const{
  return ::segno(vicino);
}

//dice di che tipo e' in {-1,0,1,2,3,4}
int vertice::tipo(int j) const {
  int ordinati[4]; for (int k=0; k<4; k++) ordinati[k]=get_vicino(k);
  ::ordina(ordinati, 3);
  int dentro=0;
  for (int i=1; i<4; i++)
    if (ordinati[i] == ordinati[i-1] && ordinati[i] == j) dentro++;
  if (dentro>2) dentro=2;
  int fuori=0;
  for (int i=1; i<4; i++)
    if (ordinati[i] == ordinati[i-1] && ordinati[i] != j) fuori++;
  int a1=ordinati[1]; int a2=ordinati[2];
  if (fuori==1 && dentro==0) return 0;  //2fuori
  if (fuori==2 && a1 != a2) return 1;   //2fuori-2fuori
  if (fuori==2 && a1 == a2) return 2;   //3fuori
  if (fuori==0 && dentro==1) return 3;  //2dentro
  if (fuori==1 && dentro==1) return 4;  //2dentro-2fuori
  return -1;                            //tutti diversi (o 4 uguali)
}

//mette in forma minima. Ritorna TRUE se era gia' in forma_minima, FALSE altrimenti.
bool grafo::set_forma_minima() {
  std::list<trasformazione> trasfmin = trasformazioni_minimizzanti();
  if (trasfmin.front() == trasformazione_banale()) return TRUE;
  trasforma(trasfmin.front());
  return FALSE;
}

//dice se e' gia' in forma minima.
bool grafo::forma_minima() const {
  std::list<trasformazione> trasfmin = trasformazioni_minimizzanti(TRUE);
  return (trasfmin.front() == trasformazione_banale());
}

//se fermati_subito==FALSE (default): restituisce tutte le trasformazioni che trasformano il grafo
//  in forma minima.
//se fermati_subito==TRUE: restituisce una trasformazione che trasforma il il grafo in un grafo
//  minore, se c'e'. Altrimenti restituisce la trasformazione banale.
//in ogni caso, restituisce una lista non vuota, e se contiene la trasformazione banale
//allora e' in testa
std::list<trasformazione> grafo::trasformazioni_minimizzanti(bool fermati_subito) const {
  const int l=lung();
  std::list<trasformazione> trasfmin = std::list<trasformazione> (1,trasformazione_banale());
  trasformazioni_minimizzanti(trasfmin, 0, fermati_subito);
  return trasfmin;
}

//funzione ricorsiva che serve a trasformazioni_minimizzanti(bool)
void grafo::trasformazioni_minimizzanti(std::list<trasformazione>& trasfmin, int liv, bool fermati_subito) const {
  static bool trovata;
  static int l;
  static std::vector<int> s;
  static std::vector<int> primi_possibili;
  if (liv==0) {
    trovata = FALSE;
    l = lung();
    s = std::vector<int>(l,0);
    
    //qui mette i vertici che possono essere in prima posizione.
    primi_possibili = std::vector<int>();
    bool trovati=FALSE; const int precedenze[6] = {4,3,2,1,0,-1};
    for (int caso=0; caso<6 && !primi_possibili.size(); caso++)
      for (int i=0; i<l; i++)
        if (tipo(i)==precedenze[caso])
          primi_possibili.push_back(i);
  }
  if (fermati_subito && trovata) return;
	if (liv<l) {
		std::vector<int> possibilita = std::vector<int>();
		if (liv==0) possibilita = primi_possibili;
		else for (int i=0; i<l; i++) if (!count(s.begin(), s.begin()+liv, i)) possibilita.push_back(i);
		for (int i=0; i<liv; i++)
			for (int quanti=1; quanti<=4; quanti++) {

				//lascia nel vettore 'possibilita' solo i vertici collegati 
				//almeno 'quanti' volte a s[i], se ce ne sono! (altrimenti non fa nulla)
				std::vector<int> possibilita_che_soddisfano = std::vector<int>();
				for (int k=0; k<possibilita.size(); k++) {
					int quanti2 = 0;
					for (int t=0; t<4; t++)
						if (get_v(s[i], t) == possibilita[k]) quanti2++;
					if (quanti2 >= quanti) 
						possibilita_che_soddisfano.push_back(possibilita[k]);
				}
				if (!possibilita_che_soddisfano.empty())
					possibilita = possibilita_che_soddisfano;

			}
		for (int t=0; t<possibilita.size(); t++) {
			s[liv]=possibilita[t];
			trasformazioni_minimizzanti(trasfmin, liv+1, fermati_subito);
    }
  }
  else {
    std::vector<int> sgiusta = perm_inversa(s);
    trasformazione trasf = trasformazione_banale();
    trasf.t[0] = sgiusta;
    grafo g = grafo(*this);
    g.trasforma(trasf);
    grafo g2 = grafo(*this);
    g2.trasforma(trasfmin.front());
    if (g < g2) {
      trasfmin = std::list<trasformazione>(1,trasf);
      trovata = TRUE;
    }
    if (g == g2)
      trasfmin.push_back(trasf);
	}
}

//serve per la classificazione dei grafi utili, cioe' a grafi(int l, TRUE)
//ci sono due indicazioni con il punto esclamativo: sono quelle che filtrano
//solo grafi irriducibili
void set_grafi(int q, int usc, int l, tensor2& w1, std::vector<int>& w2, std::ostream& file, bool solo_utili) {
  if (usc<w2[q]) usc=w2[q];
  if (q==l-1 && usc==4) {
    grafo nuovo = grafo(w1, l);
    if (nuovo.forma_minima() && (!solo_utili || nuovo.utile())) {
      if (mostra_contatore) {
				std::cout << "\rGrafi con " << l << " vertici: ";
				for (int t=0; t<2 && l>1 || t<1; t++) for (int t2=0; t2<4; t2++)
					std::cout << w1[t][t2] << " ";
				std::cout << "\r";
			}
      nuovo.salva(file);
    }
  }
  else {
    if (usc<4) 
      for (w1[q][usc]=(usc==0 || w1[q][usc-1]<q) ? q : w1[q][usc-1]; w1[q][usc]<l; w1[q][usc]++) {
        bool ok=1; w2[q]++;
        
        //5 volte lo stesso numero:
        if (++w2[w1[q][usc]]>4) ok=0;		
        
        //salto troppo in avanti:
        if (!usc && w1[q][0]>q+1)
          for (int t=q+1; t<w1[q][0]; t++) if (w2[t]==0) ok=0;
        if (usc && w1[q][usc]>w1[q][usc-1]+1)
          for (int t=w1[q][usc-1]+1; t<w1[q][usc]; t++) if (w2[t]==0) ok=0;
        
        //se il vertice 0 e' di tipo '1234' non ci devono essere catene:
        if (q>0 && w1[0][0]!=w1[0][1] && usc>0 && w1[q][usc-1]==w1[q][usc]) 
          ok=0;
				
				if (solo_utili) {

					//riccio! (e quindi non irriducibile):
					if (l>1 && w1[q][usc]==q) ok=0;
				}
		
				else {

					//riccio troppo avanti:
					if (w1[q][usc]==q && w1[0][0]!=0) ok=0;
				}

				if (ok) {
					w1[w1[q][usc]][w2[w1[q][usc]]-1]=q;
					set_grafi(q, usc+1, l, w1, w2, file, solo_utili);
				}		
				w2[q]--; w2[w1[q][usc]]--;
			}
		else {
      bool ok=1;

			//autocollegamenti devono essere in numero pari
			int quanti=0;
      for (int t=0; t<4; t++)
        if (w1[q][t]==q) quanti++;
      if (quanti % 2) ok=0;
      
      //grafo sconnesso:
      if (q<l-1 && w2[q+1]==0) ok=0;

      //w1[0][] diverso da {1,2,2,3} o {1,2,3,3}:
      if (q==0 && w1[0][1]==2 && w1[0][3]==3) ok=0;

			if (solo_utili) {
			
      //grafo non irriducibile!:
      if (l>2 && w1[q][1]==w1[q][2] && (w1[q][0]==w1[q][1] || w1[q][2]==w1[q][3]))
        ok=0;
			}

			else {
				if (q>0 && w1[q][0]==w1[q][1] && w1[q][2]==w1[q][3]) {
					
					//2f2f troppo avanti:
					if (w1[0][3]>2) ok=0;

					//2d2f troppo avanti:
					if (w1[0][0]==0 && w1[0][2]!=w1[0][3] && (w1[q][1]==q || w1[q][2]==q)) ok=0;
				}

				//3f troppo avanti
				if (w1[q][1]==w1[q][2] && (w1[q][0]==w1[q][1] || w1[q][2]==w1[q][3]) &&
						w1[0][2]!=1) ok=0;
			}

      if (ok)
        set_grafi(q+1, 0, l, w1, w2, file, solo_utili);
    }
  }
}

//serve per la classificazione dei grafi (tutti, non solo quelli utili), 
//serve cioe' a grafi(int l, FALSE)
void set_grafi(int q, int usc, int l, tensor2& w1, std::vector<int>& w2, std::ostream& file) {
  if (usc<w2[q]) usc=w2[q];
  if (q==l-1 && usc==4) {
    grafo nuovo = grafo(w1, l);
    if (nuovo.forma_minima()) {
			if (mostra_contatore) {
				std::cout << "\rGrafi con " << l << " vertici: ";
				for (int t=0; t<2 && l>1 || t<1; t++) for (int t2=0; t2<4; t2++)
					std::cout << w1[t][t2] << " ";
				std::cout << "\r";
			}
      nuovo.salva(file);
    }
  }
  else {
    if (usc<4) 
			for (w1[q][usc]=(usc==0 || w1[q][usc-1]<q) ? q : w1[q][usc-1]; w1[q][usc]<l; w1[q][usc]++) {
				bool ok=1; w2[q]++;

				//5 volte lo stesso numero:
				if (++w2[w1[q][usc]]>4) ok=0;

				//salto troppo in avanti:
				if (!usc && w1[q][0]>q+1 )
					for (int t=q+1; t<w1[q][0]; t++) if (w2[t]==0) ok=0;
				if (usc && w1[q][usc]>w1[q][usc-1]+1)
					for (int t=w1[q][usc-1]+1; t<w1[q][usc]; t++) if (w2[t]==0) ok=0;

				//riccio troppo avanti:
				if (w1[q][usc]==q && w1[0][0]!=0) ok=0;
				
				if (ok) {
					w1[w1[q][usc]][w2[w1[q][usc]]-1]=q;
					set_grafi(q, usc+1, l, w1, w2, file);
				}
				w2[q]--; w2[w1[q][usc]]--;
			}
    else {
      bool ok=1; int quanti=0;
      for (int t=0; t<4; t++)
        if (w1[q][t]==q) quanti++;

      //dispari:
      if (mod(quanti,2)==1) ok=0;
			
      //grafo sconnesso:
      if (q<l-1)
				if (w2[q+1]==0) 
					ok=0;
      
      //w[0][]= 1,2,2,3 o 1,2,3,3:
      if (q==0 && w1[0][1]==2 && w1[0][3]==3) ok=0;	

      if (q>0 && w1[q][0]==w1[q][1] && w1[q][2]==w1[q][3]) {
        
        //2f2f troppo avanti:
        if (w1[0][3]>2) ok=0;

        //2d2f troppo avanti:
        if (w1[0][0]==0 && w1[0][2]!=w1[0][3] && (w1[q][1]==q || w1[q][2]==q)) ok=0;
      }

      //3f troppo avanti
      if (w1[q][1]==w1[q][2] && (w1[q][0]==w1[q][1] || w1[q][2]==w1[q][3]) &&
          w1[0][2]!=1) ok=0;

      if (ok)
        set_grafi(q+1, 0, l, w1, w2, file);
    }
  }
}

//crea il file 'schel#.dat' con #=l, contenente 
//  se utili == FALSE (default): tutti i grafi con l vertici 
//  se utili == TRUE: tutti i grafi utili con l vertici
void set_grafi (int l, bool utili) {
  tensor2 w1(l,std::vector<int>(4,0)); std::vector<int> w2(l,0);
  char vv[14] = "schel"; strcat(vv,l); strcat(vv, ".dat");
  std::ofstream f(vv, std::ios::out);
	set_grafi (0, 0, l, w1, w2, f, utili);
  f.close();
	if (mostra_contatore) std::cout << "\n";
}

// x==y, compatibile con operator<
//ATTENZIONE: si usa questa forma, e non 
//'bool operator==(const vertice& x, const vertice& y)'
//perche' quest'ultima non puo' essere sovraccaricata,
//quindi non si comporta bene se x e y sono vertice&, ma in realta'
//ivertici od overtici
bool vertice::operator==(const vertice& y) const {
	std::vector<int> xv(4), yv(4);
	for (int t=0; t<4; t++) {xv[t] = get_vicino(t); yv[t] = y.get_vicino(t);}
  sort(xv.begin(), xv.end()); sort(yv.begin(), yv.end());
  return (xv == yv);
}

// x<y: lessicografico sulle quadruple ottenendo ordinando vicino[]
bool vertice::operator<(const vertice& y) const {
	std::vector<int> xv(4), yv(4);
	for (int t=0; t<4; t++) {xv[t] = get_vicino(t); yv[t] = y.get_vicino(t);}
  sort(xv.begin(), xv.end()); sort(yv.begin(), yv.end());
  return (xv < yv);
}

// x<y: lessicografico su (lung(), *v[])
bool grafo::operator<(const grafo& y) const {
  const int lx = lung(); const int ly = y.lung();
  if (lx<ly) return TRUE;
  if (ly<lx) return FALSE;
	for (int i=0; i<lx; i++) {
		if (*v[i] < *y[i]) return TRUE;
		if (*y[i] < *v[i]) return FALSE;
	}
	return FALSE;
}

//assegnamento
grafo& grafo::operator=(const grafo& g) {
	disalloca(v);
	v.clear();
	for (int i=0; i<g.lung(); i++)
		v.push_back(new vertice(*g[i]));
	return *this;
}

// x==y: x[] == y[];  e' compatibile con operator<
bool operator==(const grafo& x, const grafo& y) {
	const int lx=x.lung(); const int ly=y.lung();
	if (lx!=ly) return FALSE;
	for (int i=0; i<lx; i++)
		if (!(*x[i] == *y[i])) return FALSE;
	return TRUE;
}

//restituisce una permutazione banale (cioe' l'identita')
trasformazione grafo::trasformazione_banale() const {
  const int l = lung();
  trasformazione trasf;
  trasf.t.push_back(std::vector<int>(l,0));
  for (int i=0; i<l; i++) trasf.t[0][i] = i;
  return trasf;
}

void torsione::scrivi (std::ostream& f) const {
  f << std::setw(2) << t;
}

bool operator<(const torsione& t1, const torsione& t2) {
  return t1.get_t() < t2.get_t();
}

bool operator==(const torsione& t1, const torsione& t2) {
  return t1.get_t() == t2.get_t();
}

bool operator!=(const torsione& t1, const torsione& t2) {
  return t1.get_t() != t2.get_t();
}

torsione operator+(const torsione& x, const torsione& y) {
  return (x == 3 || y == 3) ? 3 : (x.t + y.t)%3;
}

torsione torsione::operator+=(const torsione& t) {
	*this = *this + t;
	return *this;
}

bool operator==(const trasformazione& t1, const trasformazione& t2) {
  return (t1.t==t2.t);
}
