#if !defined( __SURFACES_H )
#define __SURFACES_H

#include <list>
#include <iostream>

#include "defs.h"

class ografo;
class faccia;
class grafo3;
class ostream;

class vertice3{
  int vicino[3];
  int col[3];
  bool tor[3];
 public:
  vertice3();
  
  void set_vicino(int i, int j) {vicino[i]=j;}
  void set_col(int i, int j) {col[i]=j;}
  void set_tor(int i, bool j) {tor[i]=j;}

  int get_vicino(int j) const {return vicino[j];}
  int get_col(int j) const {return col[j];}
  bool get_tor(int j) const {return tor[j];}
};

class faccia_grafo3 {
  std::vector<int> passaggio;
  std::vector<int> uscita;
  std::vector<int> entrata;
 public:
  int lati() const {return passaggio.size();}
  int get_passaggio(int i) const {return passaggio[i];}
  int get_uscita(int i) const {return uscita[i];}
  int get_entrata(int i) const {return entrata[i];}
  friend std::vector<faccia_grafo3> facce_grafo3 (const grafo3& s);
};

class grafo3 {
  std::vector<vertice3> v;
  std::vector<int> colori;
  tensor3 ciclo;
  std::vector<faccia_grafo3> ff;
 public:
  grafo3(int l = 0) {v = std::vector<vertice3>(l);}
  int lung() const {return v.size();}

  int get_v(int i, int j) const {return v[i].get_vicino(j);}
  int get_col(int i, int j) const {return v[i].get_col(j);}
  bool get_tor(int i, int j) const {return v[i].get_tor(j);}
  const std::vector<int>& get_colori() const {return colori;}
  void set_col(int, int, int, int, int=0);
  void set_tor(int i, int j, bool tr);

  void set_facce();
  void scrivi_facce(std::ostream& ) const;

	bool prima_occorrenza_spigolo(int vert, int usc) const;

	std::vector<int> vertici_in_componenti_connesse() const;

  friend grafo3 grafo3_di_superficie_da_colori_facce(ografo o, bool con_cicli);
  friend grafo3 grafo3_di_bordo (const ografo&);
};

class superficie {
  std::vector<int> eulero;
	std::vector<int> orientabile;
  std::vector<int> colore;
 public:
	superficie():eulero(), orientabile(), colore() {}
  superficie(const grafo3& , int=0);
  superficie(const ografo&);
  superficie(int e, int o):colore(0), eulero(1, e), orientabile(1, o) {}

  int componenti_connesse() const {return eulero.size();}
  int get_eulero(int = -1) const;
  bool get_orientabile(int = -1) const;

	void set_eulero(int i, int e);
	void set_orientabile(int i, int o);
  void scrivi(std::ostream& ) const;

	friend bool operator==(const superficie& x, const superficie& y);
};

superficie superficie_da_colori_facce(const ografo& o);
grafo3 grafo3_di_bordo(const ografo& o);
superficie bordo(const ografo& o);
tensor2 adiacenze_bordo_su_vertice(const ografo& o);

std::list<superficie> superfici(const std::list<grafo3>& sl);
std::list<grafo3> grafi3_di_superfici_contenute (const ografo& o);
std::list<superficie> superfici_contenute(const ografo& o);

#endif
